@@ -1,10 +1,10 @@
+from __future__ import absolute_import
 from __future__ import print_function
 from __future__ import unicode_literals
 
 from argparse import ArgumentParser, SUPPRESS as SUPPRESS_ARG
 from textwrap import dedent
 from argparse import _VersionAction
-from pkg_resources import get_distribution
 from codecs import open
 import errno
 import logging
@@ -18,20 +18,23 @@ from os import stat
 import re
 import sys
 
+import ldap
 import psycopg2
 import yaml
 
 from . import __version__
-from .acl import Acl
-from .acl import process_definitions as process_acls
+from .privilege import Privilege
+from .privilege import process_definitions as process_privileges
 from .utils import (
     deepget,
     deepset,
+    iter_deep_keys,
     UserError,
     string_types,
 )
 from . import validators as V
-from .defaults import make_well_known_acls
+from .defaults import make_well_known_privileges
+from .defaults import shared_queries as default_shared_queries
 
 
 logger = logging.getLogger(__name__)
@@ -58,6 +61,7 @@ class ColoredStreamHandler(logging.StreamHandler):
     _color_map = {
         logging.DEBUG: '37',
         logging.INFO: '1;39',
+        logging.CHANGE: '1;39',
         logging.WARN: '96',
         logging.ERROR: '91',
         logging.CRITICAL: '1;91',
@@ -75,23 +79,17 @@ class ColoredStreamHandler(logging.StreamHandler):
 
 class VersionAction(_VersionAction):
     def __call__(self, parser, *a):
-        try:
-            pyldap = get_distribution('pyldap')
-        except Exception:  # pragma: nocover_py3
-            pyldap = get_distribution('python-ldap')
-
         version = (
             "%(package)s %(version)s\n"
             "psycopg2 %(psycopg2version)s\n"
-            "%(pyldap)s %(ldapversion)s\n"
+            "python-ldap %(ldapversion)s\n"
             "Python %(pyversion)s\n"
         ) % dict(
             package=__package__,
             version=__version__,
             psycopg2version=psycopg2.__version__,
             pyversion=sys.version,
-            pyldap=pyldap.project_name,
-            ldapversion=pyldap.version,
+            ldapversion=ldap.__version__,
 
         )
         print(version.strip())
@@ -124,13 +122,14 @@ def define_arguments(parser):
     )
     parser.add_argument(
         '-q', '--quiet',
-        action='store_false', dest='verbose',
-        help="hide debugging messages",
+        action='append_const', dest='verbosity', const=-1,
+        default=[V.VERBOSITIES.index(Configuration.DEFAULTS['verbosity'])],
+        help="decrease log verbosity (env: VERBOSITY)",
     )
     parser.add_argument(
         '-v', '--verbose',
-        action='store_true', dest='verbose',
-        help="add debug messages including SQL and LDAP queries (env: VERBOSE)"
+        action='append_const', dest='verbosity', const=+1,
+        help="increase log verbosity (env: VERBOSITY)"
     )
     parser.add_argument(
         '--color',
@@ -154,44 +153,51 @@ def define_arguments(parser):
     )
 
 
-def merge_acl_options(acls, acl_dict, acl_groups):
+def merge_privilege_options(privileges, acl_dict, acl_groups):
     final = dict()
     final.update(acl_dict)
     final.update(acl_groups)
-    final.update(acls)
-    return V.acls(final)
+    final.update(privileges)
+    return V.privileges(final)
 
 
-def list_unused_acl(acls, aliases):
+def list_unused_privilege(privileges, aliases):
     used = set()
     for name, aliases in aliases.items():
         if name[0] not in ('_', '.'):
             used.add(name)
             used.update(aliases)
-    unused = set(acls.keys()) - used
+    unused = set(privileges.keys()) - used
     return sorted(unused)
 
 
-def postprocess_acl_options(self, defaults=None):
+def postprocess_privilege_options(self, defaults=None):
+    # Inject default shared_queries.
+    self.setdefault('postgres', {})
+    conf_shared_queries = self['postgres'].get('shared_queries', {})
+    self['postgres']['shared_queries'] = dict(
+        default_shared_queries, **conf_shared_queries)
+
     # Compat with user defined acl_dict and acl_groups, merge in the same
     # namespace.
-    acls = defaults or {}
-    acls.update(self.pop('acls', {}))
-    acls = merge_acl_options(
-        acls,
-        self.get('acl_dict', {}),
+    privileges = defaults or {}
+    privileges.update(self.pop('acls', {}))
+    privileges.update(self.pop('privileges', {}))
+    privileges = merge_privilege_options(
+        privileges,
+        self.pop('acl_dict', {}),
         self.pop('acl_groups', {}),
     )
 
-    acls, _, self['acl_aliases'] = process_acls(acls)
+    privileges, _, self['privilege_aliases'] = process_privileges(privileges)
 
-    # Clean unused ACL starting with _ or .
-    for k in list_unused_acl(acls, self['acl_aliases']):
-        logger.debug("Drop unused hidden ACL %s", k)
-        del acls[k]
+    # Clean unused privilege starting with _ or .
+    for k in list_unused_privilege(privileges, self['privilege_aliases']):
+        logger.debug("Drop unused hidden privilege %s", k)
+        del privileges[k]
 
-    self['acl_dict'] = dict([
-        (k, Acl.factory(k, **v)) for k, v in acls.items()
+    self['privileges'] = dict([
+        (k, Privilege.factory(k, **v)) for k, v in privileges.items()
     ])
 
 
@@ -289,35 +295,52 @@ class ConfigurationError(UserError):
         )
 
 
-class NoConfigurationError(Exception):
-    pass
-
-
 def construct_yaml_str(self, node):
     # See https://stackoverflow.com/a/2967461/2613806
     return self.construct_scalar(node)
 
 
-yaml.Loader.add_constructor(u'tag:yaml.org,2002:str', construct_yaml_str)
+yaml.SafeLoader.add_constructor(u'tag:yaml.org,2002:str', construct_yaml_str)
+
+
+def check_yaml_gotchas(file_config):
+    dict_keys = ('ldap', 'postgres')
+    for key in dict_keys:
+        if key not in file_config:
+            continue
+
+        if not hasattr(file_config[key], 'items'):
+            msg = "Error ni YAML: %s: is not a dict." % key
+            raise ConfigurationError(msg)
+
+    for k, v in file_config.get('postgres', {}).items():
+        if not k.endswith('_query'):
+            continue
+        if v is None:
+            continue
+        if not v:
+            msg = "Error in YAML: postgres:%s: is empty." % k
+            raise ConfigurationError(msg)
 
 
 class Configuration(dict):
     DEFAULTS = {
         'check': False,
         'dry': True,
-        'verbose': False,
+        'verbose': None,
+        'verbosity': 'INFO',
         'color': False,
         'ldap': {
-            'uri': '',
-            'host': '',
-            'port': 389,
-            'binddn': '',
+            'uri': None,
+            'host': None,
+            'port': None,
+            'binddn': None,
             'user': None,
-            'password': '',
+            'password': None,
+            'referrals': None,
         },
         'postgres': {
             'dsn': '',
-            'blacklist': ['pg_*', 'postgres'],
             'databases_query': dedent("""\
             SELECT datname FROM pg_catalog.pg_database
             WHERE datallowconn IS TRUE ORDER BY 1;
@@ -328,12 +351,14 @@ class Configuration(dict):
             # https://www.postgresql.org/docs/current/static/catalog-pg-auth-members.html
             'roles_query': dedent("""\
             SELECT
-              role.rolname, array_agg(members.rolname) AS members, {options}
+              role.rolname, array_agg(members.rolname) AS members,
+              {options},
+              pg_catalog.shobj_description(role.oid, 'pg_authid') as comment
             FROM
               pg_catalog.pg_roles AS role
             LEFT JOIN pg_catalog.pg_auth_members ON roleid = role.oid
             LEFT JOIN pg_catalog.pg_roles AS members ON members.oid = member
-            GROUP BY role.rolname, {options}
+            GROUP BY role.rolname, {options}, comment
             ORDER BY 1;
             """),
             'owners_query': dedent("""\
@@ -343,11 +368,16 @@ class Configuration(dict):
             ORDER BY 1;
             """),
             'managed_roles_query': None,
+            'roles_blacklist_query': [
+                'pg_*', 'postgres', 'rds_*', 'rds*admin',
+            ],
             'schemas_query': dedent("""\
             SELECT nspname FROM pg_catalog.pg_namespace
             ORDER BY 1;
             """),
+            'shared_queries': {},
         },
+        'privileges': {},
         'acls': {},
         'acl_dict': {},
         'acl_groups': {},
@@ -358,25 +388,30 @@ class Configuration(dict):
         Mapping('color'),
         Mapping('check'),
         Mapping('dry'),
-        Mapping('verbose', env=['VERBOSE', 'DEBUG']),
+        Mapping('verbose', env=[]),
+        Mapping('verbosity', processor=V.verbosity),
         Mapping('ldap:uri'),
         Mapping('ldap:host'),
         Mapping('ldap:port'),
         Mapping('ldap:binddn', env=['LDAPBINDDN', 'LDAP_BIND']),
         Mapping('ldap:user'),
         Mapping('ldap:password', secret=True),
+        Mapping('ldap:referrals'),
         Mapping(
             'postgres:dsn', env='PGDSN',
             secret=r'(?:password=|:[^/][^/].*@)',
         ),
-        Mapping('postgres:blacklist', env=None),
         Mapping('postgres:databases_query', env=None),
         Mapping('postgres:owners_query', env=None),
         Mapping('postgres:roles_query', env=None),
         Mapping('postgres:managed_roles_query', env=None),
+        Mapping('postgres:roles_blacklist_query', env=None),
         Mapping('postgres:schemas_query', env=None),
-        Mapping('acls', env=None, processor=V.acls),
-        Mapping('acl_dict', processor=V.acldict),
+        Mapping(
+            'postgres:shared_queries', processor=V.shared_queries, env=None),
+        Mapping('privileges', env=None, processor=V.privileges),
+        Mapping('acls', env=None, processor=V.privileges),
+        Mapping('acl_dict', processor=V.privileges),
         Mapping('acl_groups', env=None),
         Mapping('sync_map', env=None, processor=V.syncmap)
     ]
@@ -386,7 +421,7 @@ class Configuration(dict):
 
     _file_candidates = [
         './ldap2pg.yml',
-        '~/.config/lda2pg.yml',
+        '~/.config/ldap2pg.yml',
         '/etc/ldap2pg.yml',
     ]
 
@@ -414,13 +449,14 @@ class Configuration(dict):
                 return os.path.realpath(candidate), stat_.st_mode
             except OSError as e:
                 if e.errno == errno.EACCES:
-                    logger.warn("Can't read %s: permission denied.", candidate)
+                    logger.warning(
+                        "Can't read %s: permission denied.", candidate)
 
         if custom:
             message = "Can't access configuration file %s." % (custom,)
             raise UserError(message, exit_code=os.EX_NOINPUT)
         else:
-            raise NoConfigurationError("No configuration file found")
+            raise ConfigurationError("No configuration file found")
 
     EPILOG = dedent("""\
 
@@ -434,6 +470,21 @@ class Configuration(dict):
     def has_ldap_query(self):
         return [m['ldap'] for m in self['sync_map'] if 'ldap' in m]
 
+    def bootstrap(self, environ=os.environ):
+        debug = environ.get('DEBUG', '').lower() in ('1', 'y')
+        verbose = debug or environ.get('VERBOSE', '').lower() in ('1', 'y')
+        verbosity = environ.get('VERBOSITY', 'DEBUG' if verbose else 'INFO')
+
+        self['debug'] = debug
+        try:
+            self['verbosity'] = V.verbosity(verbosity)
+        except ValueError as e:
+            raise UserError('Failed to boostrap: %s.' % (e,))
+        self['color'] = sys.stderr.isatty()
+
+        dictConfig(self.logging_dict())
+        return debug
+
     def load(self, argv=None):
         # argv processing.
         logger.debug("Processing CLI arguments.")
@@ -442,56 +493,58 @@ class Configuration(dict):
             # Only store value from argv. Defaults are managed by
             # Configuration.
             argument_default=SUPPRESS_ARG,
-            description="PostgreSQL roles and ACL management.",
+            description="PostgreSQL roles and privileges management.",
             epilog=self.EPILOG,
         )
         define_arguments(parser)
         args = parser.parse_args(sys.argv[1:] if argv is None else argv)
 
-        if hasattr(args, 'verbose') or hasattr(args, 'color'):
-            # Switch to verbose before loading file.
-            self['verbose'] = getattr(args, 'verbose', self['verbose'])
-            self['color'] = getattr(args, 'color', self['color'])
-            dictConfig(self.logging_dict())
+        # Setup logging before parsing options. Reset verbosity with env var,
+        # and compute verbosity from cumulated args.
+        args.verbosity[0] = V.VERBOSITIES.index(self['verbosity'])
+        self['verbosity'] = V.verbosity(args.verbosity)
+        if hasattr(args, 'color'):
+            self['color'] = args.color
+        dictConfig(self.logging_dict())
 
         logger.info("Starting ldap2pg %s.", __version__)
 
         # File loading.
-        try:
-            filename, mode = self.find_filename(os.environ, args)
-        except NoConfigurationError:
-            logger.debug("No configuration file found.")
-            file_config = {}
+        filename, mode = self.find_filename(os.environ, args)
+        if filename == '-':
+            logger.info("Reading configuration from stdin.")
+            file_config = self.read(sys.stdin, 'stdin', mode)
         else:
-            if filename == '-':
-                logger.info("Reading configuration from stdin.")
-                file_config = self.read(sys.stdin, 'stdin', mode)
-            else:
-                logger.info("Using %s.", filename)
-                try:
-                    with open(filename, encoding='utf-8') as fo:
-                        file_config = self.read(fo, filename, mode)
-                except OSError as e:
-                    msg = "Failed to read configuration: %s" % (e,)
-                    raise UserError(msg)
+            logger.info("Using %s.", filename)
+            try:
+                with open(filename, encoding='utf-8') as fo:
+                    file_config = self.read(fo, filename, mode)
+            except OSError as e:
+                msg = "Failed to read configuration: %s" % (e,)
+                raise UserError(msg)
+
+        V.alias(
+            file_config.get('postgres', {}),
+            'roles_blacklist_query', 'blacklist',
+        )
 
         # Now close stdin. To make SASL non-interactive.
         if not self.get('debug'):
             sys.stdin.close()
 
+        check_yaml_gotchas(file_config)
+        self.warn_unknown_config(file_config)
+
         # Now merge all config sources.
-        acl_defaults = make_well_known_acls()
+        default_privileges = make_well_known_privileges()
         try:
             self.merge(file_config=file_config, environ=os.environ, args=args)
-            postprocess_acl_options(self, acl_defaults)
+            postprocess_privilege_options(self, default_privileges)
         except ValueError as e:
             raise ConfigurationError("Failed to load configuration: %s" % (e,))
 
         logger.debug("Configuration loaded.")
 
-        if not self['sync_map']:
-            logger.warn("Empty synchronization map!")
-
     def merge(self, file_config, environ=os.environ, args=object()):
         for mapping in self.MAPPINGS:
             value = mapping.process(
@@ -502,23 +555,30 @@ class Configuration(dict):
             )
             deepset(self, mapping.path, value)
 
+        if self['verbose'] is not None:
+            self['verbosity'] = 'DEBUG' if self['verbose'] else 'INFO'
+
     def read(self, fo, name, mode):
         try:
-            payload = yaml.load(fo) or {}
+            payload = yaml.safe_load(fo)
         except yaml.error.YAMLError as e:
             msg = "YAML error with %s: %s" % (name, e)
             raise ConfigurationError(msg)
 
+        if payload is None:
+            raise ConfigurationError("Configuration is empty.")
         if isinstance(payload, list):
             payload = dict(sync_map=payload)
         if not isinstance(payload, dict):
             raise ConfigurationError("Configuration file must be a mapping.")
+        if 'sync_map' not in payload:
+            raise ConfigurationError("sync_map configuration is required.")
+
         payload['world_readable'] = bool(mode & 0o077)
         return payload
 
     def logging_dict(self):
-        formatter = 'verbose' if self['verbose'] else 'info'
-
+        formatter = 'verbose' if self['verbosity'] == 'DEBUG' else 'info'
         return {
             'version': 1,
             'formatters': {
@@ -547,7 +607,17 @@ class Configuration(dict):
             },
             'loggers': {
                 __package__: {
-                    'level': 'DEBUG' if self['verbose'] else 'INFO',
+                    'level': self['verbosity'],
                 },
             },
         }
+
+    def warn_unknown_config(self, config):
+        known_keys = set([m.path for m in self.MAPPINGS] + ['world_readable'])
+
+        for k in iter_deep_keys(config):
+            if k.startswith('privileges'):
+                continue
+
+            if k not in known_keys:
+                logger.warning("Unknown config entry: %s.", k)
